#!python3
#
# meli - scripts/markdown_doc_lints.py
#
# Copyright 2024 Manos Pitsidianakis
#
# This file is part of meli.
#
# meli is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# meli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with meli. If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: EUPL-1.2 OR GPL-3.0-or-later

""" Format and validate with:

```
black scripts/markdown_doc_lints.py
pylint scripts/markdown_doc_lints.py
mypy --strict-equality -strict scripts/markdown_doc_lints.py
```
"""

# pylint: disable=invalid-name, redefined-outer-name

from typing import Mapping, Any, Tuple
import io
import tomllib
import argparse
import sys
import warnings
import re


def validate_cargo_toml(cargo_toml: io.BufferedReader) -> Mapping[str, Any]:
    """Parse Cargo.toml contents and return it as a dictionary"""
    try:
        cargo_toml_dict: Mapping[str, Any] = tomllib.load(cargo_toml)
    except tomllib.TOMLDecodeError as exc:
        print(cargo_toml.name, "could not be parsed as TOML")
        raise exc

    if "features" not in cargo_toml_dict:
        print(
            "`features` key not found in",
            cargo_toml.name,
            "is it the correct Cargo.toml you wanted to pass?",
        )
        sys.exit(1)

    return cargo_toml_dict


def validate_default_features(
    readme_md: io.TextIOWrapper, readme_md_contents: str, features: dict[str, list[str]]
) -> bool:
    """Ensure README.md has a part mentioning the default features in a toml block"""
    default_features_md_paragraph_re = r"```toml\ndefault\s*=.*?\n```"
    if "default" in features:
        default_features_list: list[str] = list(features["default"])
        default_feature_string = """```toml
default = ["""
        for f in default_features_list:
            if not default_feature_string.endswith(" ["):
                default_feature_string += ", "
            default_feature_string += f'"{f}"'
        default_feature_string += """]
```
"""
        if default_feature_string not in readme_md_contents:
            existing_match = re.search(
                default_features_md_paragraph_re, readme_md_contents, re.IGNORECASE
            )
            if existing_match is None:
                print(
                    "A paragraph mentioning the default contents must be present in",
                    readme_md.name,
                )
                print("Add the following markdown:")
            else:
                print(
                    "The paragraph in",
                    readme_md.name,
                    "mentioning the default contents has invalid or incorrectly formatted contents",
                )
                print("What exists in", readme_md.name, "is:")
                print(existing_match.group(0))
                print("Replace it with the following markdown:")
            print(default_feature_string, end="")
            return False
    return True


def validate_valid_features_mentioned(
    cargo_toml: io.BufferedReader,
    readme_md: io.TextIOWrapper,
    readme_md_contents: str,
    features_list: list[str],
) -> bool:
    """Ensure that valid features as parsed from Cargo.toml manifest are
    mentioned in provided README.md with the correct format (a named HTML anchor element)
    """
    missing: list[Tuple[str, str]] = []
    for feat in features_list:
        anchor_fmt = f'<a name="{feat}-feature">`{feat}`</a>'
        if anchor_fmt not in readme_md_contents:
            missing.append((feat, anchor_fmt))
    if len(missing) == len(features_list):
        print("None of the features in", cargo_toml.name, "appear in", readme_md.name)
        print(f"{features_list=}")
        return False
    if len(missing) > 0:
        print(
            "The following",
            len(missing),
            "feature" if len(missing) == 1 else "features",
            "are not mentioned in",
            readme_md.name,
        )
        print("")
        for m, anchor in missing:
            print("-", m, "add the following string in the Features section:", anchor)
        return False
    return True


def validate_readme_feature_named_anchors(
    cargo_toml: io.BufferedReader,
    readme_md_contents: str,
    features_list: list[str],
) -> bool:
    """Ensure name attribute of anchor HTML elements match the feature name,
    and also that all such elements refer to valid features
    (as parsed from the Cargo.toml manifest)"""
    anchor_re = (
        r"[<]a name=[\"](?P<name_attr>.+)-feature[\"][>][`](?P<name>[^`]+)[`][<].a[>]"
    )
    matches = re.finditer(anchor_re, readme_md_contents)
    newline_re = re.compile(r"\n")
    retval = True
    for match in matches:
        if match.group("name_attr") != match.group("name"):
            start_line = (
                len(newline_re.findall(readme_md_contents, 0, match.start("name_attr")))
                + 1
            )
            end_line = (
                len(newline_re.findall(readme_md_contents, 0, match.end("name_attr")))
                + 1
            )
            if start_line == end_line:
                print("In line ", start_line, ",", end="", sep="")
            else:
                print("In lines ", start_line, ":", end_line, ",", end="", sep="")
            print(
                " attribute name `",
                match.group("name_attr"),
                "` does not match feature name `",
                match.group("name"),
                "`",
                sep="",
            )
            retval = False
        for feature_name in ["name_attr", "name"]:
            if feature_name == "name" and match.group(feature_name) == match.group(
                "name_attr"
            ):
                continue
            if (
                len(match.group(feature_name)) > 0
                and match.group(feature_name) not in features_list
            ):
                start_line = (
                    len(
                        newline_re.findall(
                            readme_md_contents, 0, match.start(feature_name)
                        )
                    )
                    + 1
                )
                end_line = (
                    len(
                        newline_re.findall(
                            readme_md_contents, 0, match.end(feature_name)
                        )
                    )
                    + 1
                )
                if start_line == end_line:
                    print("In line ", start_line, ",", end="", sep="")
                else:
                    print("In lines ", start_line, ":", end_line, ",", end="", sep="")
                print(
                    " feature `",
                    match.group(feature_name),
                    "` is mentioned but it does not appear in the features defined in ",
                    cargo_toml.name,
                    sep="",
                )
                retval = False
    return retval


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Parses a README.md and the crate's Cargo.toml"
        "to ensure all compile-time features are documented. Exits with 1 if an error was found."
        "\n"
        "Example usage:\n"
        "python3 scripts/markdown_doc_lints.py melib/README.md melib/Cargo.toml"
    )
    parser.add_argument("README.md", type=argparse.FileType("r"), help="README.md file")
    parser.add_argument(
        "Cargo.toml", type=argparse.FileType("rb"), help="Cargo.toml manifest file"
    )
    args = parser.parse_args()

    readme_md: io.TextIOWrapper = vars(args)["README.md"]
    cargo_toml: io.BufferedReader = vars(args)["Cargo.toml"]

    if not readme_md.name.endswith("README.md"):
        warnings.warn(
            "The file given as the README.md argument is not named README.md. Was this a mistake?"
        )

    readme_md_contents: str = vars(args)["README.md"].read()

    cargo_toml_dict: Mapping[str, Any] = validate_cargo_toml(cargo_toml)

    features: dict[str, list[str]] = cargo_toml_dict["features"]

    success = True

    success &= validate_default_features(readme_md, readme_md_contents, features)

    for to_remove in ["default", "debug-tracing"]:
        if to_remove in features:
            del features[to_remove]

    features_list: list[str] = list(features)

    success &= validate_valid_features_mentioned(
        cargo_toml, readme_md, readme_md_contents, features_list
    )

    success &= validate_readme_feature_named_anchors(
        cargo_toml, readme_md_contents, features_list
    )

    sys.exit(0 if success else 1)



# meli - scripts/make_html_manual_page.py
#
# Copyright 2023 Manos Pitsidianakis
#
# This file is part of meli.
#
# meli is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# meli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with meli. If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import subprocess
from threading import Timer
from html.parser import HTMLParser
import argparse
import sys
import re
import shutil
import textwrap
from subprocess import PIPE
from urllib.parse import urlparse
from pathlib import Path
import http.client
from http import HTTPMethod, HTTPStatus
import signal
import functools
from bs4 import BeautifulSoup

NO_TTY = False


def clear_line(signum, frame):
    if NO_TTY:
        return
    columns = shutil.get_terminal_size().columns
    message = chr(27) + "[0G"  # go to start of line
    sys.stdout.write(message)
    message = " " * columns + "\r"
    sys.stdout.write(message)
    sys.stdout.flush()
    draw_progress.max_cols = 0


signal.signal(signal.SIGWINCH, clear_line)

MIRRORS = [
    "http://man7.org/linux/man-pages/man%S/%N.%S.html",
    "http://manpages.debian.org/stable/%N.%S.en.html",
    "http://man.bsd.lv/%N.%S",
    "http://man.archlinux.org/man/%N.%S",
    "http://man.voidlinux.org/%N.%S",
    "http://man.bsd.lv/OpenBSD-7.0/%N.%S",
    "http://man.bsd.lv/FreeBSD-13.0/%N.%S",
    "http://man.bsd.lv/POSIX-2013/%N.%S",  # last resorts
    "http://man.bsd.lv/UNIX-7/%N.%S",
    "https://www.unix.com/man-page/mojave/%S/%N/",
    "http://linux.die.net/man/%S/%N",
]


def add_progress(count=1):
    add_progress.count += count


add_progress.count = 0.0


def draw_progress(total, count=None, status=""):
    if NO_TTY:
        return
    columns = shutil.get_terminal_size().columns
    status = textwrap.shorten(
        status, width=columns - draw_progress.bar_len - len("100.0% ...") - 8
    )
    if count is None:
        count = add_progress.count
    bar_len = draw_progress.bar_len
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = "=" * filled_len + "-" * (bar_len - filled_len)

    message = f"[{bar}] {percents}% ...{status}"
    draw_progress.max_cols = max(len(message) + 1, draw_progress.max_cols)
    spaces = " " * (draw_progress.max_cols - len(message))
    message += f"{spaces}\r"
    sys.stdout.write(message)
    sys.stdout.flush()


draw_progress.max_cols = 0
draw_progress.bar_len = 62


class ManFixer(HTMLParser):
    whitespace = r"\s{2,}"
    output = ""
    extract_href = False

    def reset(self):
        self.output = ""
        super().reset()

    def handle_starttag(self, tag, attrs):
        attrs = {a[0]: a[1] for a in attrs}
        if tag == "a" and self.extract_href and "href" in attrs:
            self.output += re.sub(self.whitespace, " ", attrs["href"]).replace(
                "\ufeff", ""
            )
            self.output += " "

    def handle_endtag(self, tag):
        pass

    def handle_data(self, data):
        self.output += re.sub(self.whitespace, " ", data).replace("\ufeff", "")

    @staticmethod
    def extract(input_):
        parser = ManFixer()
        parser.feed(input_)
        return parser.output


@functools.cache
def give_me_head(url_):
    o = urlparse(url_)
    conn = http.client.HTTPSConnection(o.hostname, timeout=6)
    conn.request(HTTPMethod.HEAD, o.path)
    response = conn.getresponse()
    if response.status in (
        HTTPStatus.FOUND,
        HTTPStatus.TEMPORARY_REDIRECT,
        HTTPStatus.PERMANENT_REDIRECT,
        HTTPStatus.MOVED_PERMANENTLY,
    ):
        # print("for ", url_, "following redirect", response.status)
        give_me_head.redirects += 1
        if give_me_head.redirects > 3:
            return None
        if response.getheader("Location"):
            # print("for ", url_, "following redirect to ", response.getheader("Location"))
            return give_me_head(response.getheader("Location"))
        print("bailout")
        return None
    # print("for ", url_, "code is ", response.status)
    give_me_head.redirects = 0
    if response.status == http.HTTPStatus.OK:
        return url_
    return None


give_me_head.redirects = 0


def man_to_path(man: str) -> str:
    exp = r"(.+)(\d{1,})$"
    result = re.match(exp, man)
    if not result:
        return man
    return f"{result[2]}/{result[1][:-1]}"


def draw_spinner():
    if NO_TTY:
        return
    message = (
        chr(27)
        + "["
        + str(draw_spinner.columns - 1)
        + "C"
        + draw_spinner.frames[draw_spinner.index]
        + "\r"
    )
    sys.stdout.write(message)
    sys.stdout.flush()
    draw_spinner.index += 1
    draw_spinner.index = draw_spinner.index % len(draw_spinner.frames)
    draw_spinner.timer = Timer(draw_spinner.interval, draw_spinner)
    draw_spinner.timer.start()


draw_spinner.interval = 0.1
draw_spinner.columns = shutil.get_terminal_size().columns
draw_spinner.frames = ["|", "/", "-", "\\"]
draw_spinner.index = 0

if __name__ == "__main__":
    draw_spinner.timer = Timer(draw_spinner.interval, draw_spinner)
    parser = argparse.ArgumentParser(
        description="Generates a <table> element from a mdoc manpage."
    )
    parser.add_argument("page", type=str, help="mdoc file")
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        help="name of output file",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--name",
        "-n",
        type=str,
        help="name used for html IDs. defaults to file name stem.",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--refs",
        type=bool,
        help="find external manpages and hyperlink to them",
        required=False,
        default=True,
    )
    parser.add_argument(
        "--no-tty",
        help="don't draw progress animation",
        required=False,
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--include-refs",
        type=str,
        help="comma separated list of manpages to relatively hyperlink",
        required=False,
        default="",
    )
    parser.add_argument(
        "--exclude-refs",
        type=str,
        help="comma separated list of manpages to not hyperlink",
        required=False,
        default="",
    )
    parser.add_argument(
        "--mandoc",
        type=str,
        help="alternative mandoc binary path",
        required=False,
        default="mandoc",
    )
    parser.add_argument(
        "--no-css",
        help="don't prepend <style> element",
        required=False,
        action="store_true",
        default=False,
    )

    args = parser.parse_args()
    if args.exclude_refs:
        args.exclude_refs = [s.strip() for s in args.exclude_refs.split(",")]
    if args.include_refs:
        args.include_refs = [s.strip() for s in args.include_refs.split(",")]
    if not args.output:
        args.output = Path.cwd() / (Path(args.page).name + ".html")
    if not args.name:
        args.name = Path(args.page).name

    manpage = None
    with open(args.page, "r", encoding="utf-8") as f:
        manpage = f.read()
    if args.refs:
        refs_url = ',man="%N\t%S"'
    else:
        refs_url = ""

    NO_TTY = args.no_tty
    html_output = subprocess.run(
        f'{args.mandoc} -I os="rendered by mandoc" -Kutf-8 -Ofragment,toc,includes="#%I"{refs_url} -Thtml "{args.page}" | sed \'s/\\s*<\\/pre/<\\/pre/\'',
        stdout=PIPE,
        shell=True,
        check=True,
    ).stdout.decode("utf-8")

    html_output = re.sub(
        r"(?:(?:[⟨])|(?:&#x27E8;))(.+)(?:(?:[⟩])|(?:&#x27E9;))",
        '<kbd class="manpage-kbd">\\1</kbd>',
        html_output,
        flags=re.MULTILINE,
    )

    soup = BeautifulSoup(html_output, "html.parser")
    targets = set()
    for target in soup.find_all(lambda tag: tag.has_attr("id")):
        id_ = target.get("id")
        targets.add(id_)
    root_table = next(soup.children)
    root_table["id"] = args.name

    if args.refs:
        total = len(soup.find_all("a"))
        print(f"Replacing `href` attributes in {total} hyperlinks...")

        draw_spinner.timer.start()
        for link in soup.find_all("a"):
            href = link.get("href")
            if href.startswith("#") and href[1:] in targets:
                link["href"] = "#" + args.name + "_" + href[1:]
                add_progress()
                draw_progress(total)
            elif href.startswith("#"):
                add_progress()
                draw_progress(total)
            else:
                exp = r"(.+)\t(.+)$"
                result = re.match(exp, href)
                if result:
                    link["href"] = f"./{result[1]}.{result[2]}.html"
                    name = result[1]
                    section = result[2]
                    if (
                        name in args.exclude_refs
                        or f"{name}.{section}" in args.exclude_refs
                    ):
                        add_progress()
                        draw_progress(
                            total,
                            status=f"{name}.{section}: Excluding ref because it is in --exclude-refs list. Leaving it as {link['href']}",
                        )
                        continue
                    found = False
                    for url in MIRRORS:
                        add_progress(1.0 / (len(MIRRORS) * 1.0))
                        if found:
                            continue
                        draw_progress(
                            total,
                            status=f"{name}.{section}: searching for an online mirror",
                        )
                        url_ = url.replace("%N", name).replace("%S", section)
                        try:
                            got = give_me_head(url_)
                            if got:
                                link["href"] = got
                                found = True
                                continue
                        except Exception as exc:
                            if "handshake operation timed out" not in str(exc):
                                print(f"got {exc} for url {url_}")
                    if (
                        not found
                        and args.include_refs
                        and (
                            name in args.include_refs
                            or f"{name}.{section}" in args.include_refs
                        )
                    ):
                        link["href"] = ""
                        draw_progress(
                            total,
                            status=f"{name}.{section}: Excluding ref because it was not found online and is not in --include-refs list. Leaving it empty.",
                        )
                else:
                    add_progress()
                    draw_progress(total)
        draw_spinner.timer.cancel()
        clear_line(None, None)

    for target in soup.find_all(lambda tag: tag.has_attr("id")):
        id_ = target.get("id")
        if id_ in targets:
            id_ = args.name + "_" + id_
            target["id"] = id_

    with open(args.output, "w", encoding="utf-8") as f:
        if not args.no_css:
            f.write(
                """
            <style>
            code.Ic, code.Li, code.Cm, code.Nm, kbd.manpage-kbd{
                display: inline-block;
            }
            kbd {
              background-color: #eee;
              border-radius: 3px;
              border: 1px solid #b4b4b4;
              box-shadow:
                0 1px 1px rgba(0, 0, 0, 0.2),
                0 2px 0 0 rgba(255, 255, 255, 0.7) inset;
              color: #333;
              display: inline-block;
              font-size: 0.85em;
              font-weight: 700;
              line-height: 1;
              padding: 2px 4px;
              white-space: nowrap;
            }
            code {
              background-color: #eee;
              border-radius: 3px;
              font-family: courier, monospace;
              padding: 0 3px;
            }
            </style>
            """
            )
        f.write(str(soup))
    print("Written to ", args.output)



# meli - scripts/make_html_manual_page.py
#
# Copyright 2023 Manos Pitsidianakis
#
# This file is part of meli.
#
# meli is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# meli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with meli. If not, see <http://www.gnu.org/licenses/>.


"""
This script generates MDOC(7) bibliographic ("reference") blocks from a
hardcoded list of RFC ids. They can be used for inclusion in a manual page.

The output is formatted like this example from mdoc(7):


 Rs
     Begin a bibliographic (“reference”) block. Does not have any head
     arguments. The block macro may only contain %A, %B, %C, %D, %I, %J, %N,
     %O, %P, %Q, %R, %T, %U, and %V child macros (at least one must be
     specified).

     Examples:

     .Rs
     .%A J. E. Hopcroft
     .%A J. D. Ullman
     .%B Introduction to Automata Theory, Languages, and Computation
     .%I Addison-Wesley
     .%C Reading, Massachusetts
     .%D 1979
     .Re

     If an Rs block is used within a SEE ALSO section, a vertical space is
     asserted before the rendered output, else the block continues on the
     current line.

For example, the output for RFC1524 which describes the `mailcap` format is:

  .Rs
  .%B RFC1524 A User Agent Configuration Mechanism For Multimedia Mail Format Information
  .%I Legacy
  .%D September 01, 1993
  .%A Dr. Nathaniel S. Borenstein
  .%U https://datatracker.ietf.org/doc/rfc1524/
  .Re
"""

from urllib.parse import urlparse
import json
from datetime import datetime
import http.client
from http import HTTPMethod, HTTPStatus
import functools
import argparse
import re
import time
import random
import sys


@functools.cache
def get_request(url_):
    """Perform an HTTP GET request, follow any redirections and return the final HTTP content."""
    o = urlparse(url_)
    conn = http.client.HTTPSConnection(o.hostname, timeout=6)
    conn.request(HTTPMethod.GET, o.path)
    response = conn.getresponse()
    if response.status in (
        HTTPStatus.FOUND,
        HTTPStatus.TEMPORARY_REDIRECT,
        HTTPStatus.PERMANENT_REDIRECT,
        HTTPStatus.MOVED_PERMANENTLY,
    ):
        get_request.redirects += 1
        if get_request.redirects > 3:
            return None
        if response.getheader("Location"):
            return get_request(response.getheader("Location"))
        return None
    get_request.redirects = 0
    if response.status == http.HTTPStatus.OK:
        return response.read()
    return None


get_request.redirects = 0


@functools.cache
def head_request(url_):
    """Perform an HTTP HEAD request, follow any redirections and return the final URL."""
    o = urlparse(url_)
    conn = http.client.HTTPSConnection(o.hostname, timeout=6)
    conn.request(HTTPMethod.HEAD, o.path)
    response = conn.getresponse()
    if response.status in (
        HTTPStatus.FOUND,
        HTTPStatus.TEMPORARY_REDIRECT,
        HTTPStatus.PERMANENT_REDIRECT,
        HTTPStatus.MOVED_PERMANENTLY,
    ):
        head_request.redirects += 1
        if head_request.redirects > 3:
            return None
        if response.getheader("Location"):
            return head_request(response.getheader("Location"))
        return None
    head_request.redirects = 0
    if response.status == http.HTTPStatus.OK:
        return url_
    return None


head_request.redirects = 0


def fetch_rfc_json(ident: str) -> str:
    """Fetch JSON source for an "rfcNNNN" from IETF's Datatracker."""
    url = f"https://datatracker.ietf.org/doc/{ident.lower()}/doc.json"
    body = get_request(url)
    return json.loads(body)


def rfc_to_citation(ident: str) -> str:
    """Convert an "rfcNNNN" input to an mdoc(7) citation block"""
    # Fetch json as a dictionary
    response_dict = fetch_rfc_json(ident)
    url_path = f"/doc/{ident}/"
    retval = ".Rs\n"
    retval += f".%B {response_dict['name'].upper()} {response_dict['title']}\n"
    retval += f".%I {response_dict['stream']}\n"
    dateval = datetime.strptime(response_dict["time"].split(" ")[0], "%Y-%m-%d")
    for rev in response_dict["rev_history"]:
        if rev["name"] != ident:
            continue
        dateval = datetime.strptime(rev["published"].split("T")[0], "%Y-%m-%d")
        if "url" in rev:
            url_path = rev["url"]
    retval += f".%D {dateval.strftime('%B %d, %Y')}\n"
    if "authors" in response_dict:
        for author in response_dict["authors"]:
            if "name" not in author:
                continue
            retval += f".%A {author['name']}\n"
    doc_url = head_request(f"https://datatracker.ietf.org{url_path}")
    if doc_url:
        retval += f".%U {doc_url}\n"
    retval += ".Re"
    return retval


DEFAULT_RFCS = [
    "rfc1524",
    "rfc2047",
    "rfc2183",
    "rfc2369",
    "rfc2426",
    "rfc3156",
    "rfc3461",
    "rfc3501",
    "rfc3676",
    "rfc3691",
    "rfc3977",
    "rfc4549",
    "rfc4616",
    "rfc4954",
    "rfc5321",
    "rfc5322",
    "rfc6048",
    "rfc6152",
    "rfc6350",
    "rfc6532",
    "rfc6868",
    "rfc7162",
    "rfc8620",
    "rfc8621",
]

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            'Generates MDOC(7) bibliographic ("reference") blocks from '
            "a hardcoded list of RFC ids. They can be used for inclusion in a manual "
            "page. If no RFC id arguments are given, a default hard-coded list is "
            "processed instead."
        ),
    )

    def rfc_title(string):
        """Validate `string` as an rfc id number, with a case-insensive "RFC" prefix"""
        if re.search(r"^[rR][fF][cC]\d{3,5}$", string, re.MULTILINE):
            return string.lower()
        raise ValueError(f"{string} is not of the form 'rfcNNNN' (case insensitive)")

    parser.add_argument(
        "-o",
        "--output",
        type=argparse.FileType("w"),
        default=sys.stdout,
        help="Output file. Default is stdout",
    )
    parser.add_argument(
        metavar="RFC_TITLE",
        dest="rfc_list",
        nargs="*",
        type=rfc_title,
        default=DEFAULT_RFCS,
        help="space-seperated RFC titles of the form RFCNNNN",
    )
    args = parser.parse_args()
    for rfc in args.rfc_list:
        print(rfc_to_citation(rfc), file=args.output)
        time.sleep(random.uniform(0, 2))
